/********************************************************************
 *                                                                  *
 * HAV Library, Copyright 1994-2004 Gromada.com                     *
 *                                                                  *
 * THIS FILE IS PART OF THE "HAV Library" SOURCE CODE.              *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS LIBRARY SOURCE IS     *
 * GOVERNED BY A BSD-STYLE SOURCE LICENSE INCLUDED WITH THIS SOURCE *
 * IN 'LICENSE.TXT'. PLEASE READ THESE TERMS BEFORE DISTRIBUTING.   *
 *                                                                  *
 ********************************************************************
 *                                                                  *
 * hav.h - common data types and function protoypes                 *
 *                                                                  *
 ********************************************************************/


#ifndef  _HAV_H_
#define  _HAV_H_


#include  <stdio.h>

#include  "gromada/types.h"
#include  "gromada/file.h"


#ifdef  __cplusplus
extern "C" {
#endif

//#pragma pack(1)


// format/packet IDs
#define  HAV_VER_MAJOR   1
#define  HAV_VER_MIDDLE  0
#define  HAV_VER_MINOR   0
#define  HAV_VER  			 ((((HAV_VER_MAJOR << 16) + HAV_VER_MIDDLE) << 8) + HAV_VER_MINOR)
#define  HAV_VER_STR     "1.0.0"  // format version
#define  HAV_ID          "HAV\0"
#define  HAV_VIDEO_ID    "VID\0"
#define  HAV_AUDIO_ID    "AUD\0"
#define  HAV_PALETTE_ID  "PAL\0"
// size of headers used in HAV format
#define  HAV_HEADER_SIZE          16
#define  HAV_VIDEO_HEADER_SIZE    (HAV_HEADER_SIZE + 20)
#define  HAV_PALETTE_HEADER_SIZE  (HAV_HEADER_SIZE + 8)
#define  HAV_AUDIO_HEADER_SIZE    (HAV_HEADER_SIZE + 12)
// file position table
#define  HAV_FPT_ENTRIES  64
#define  HAV_FPT_SIZE     (sizeof(Int64) * HAV_FPT_ENTRIES + sizeof(Int64))
// palette size
#define  HAV_PALETTE_ENTRY_SIZE  3
#define  HAV_PALETTE_ENTRIES     256
#define  HAV_PALETTE_SIZE        (HAV_PALETTE_ENTRY_SIZE * HAV_PALETTE_ENTRIES)
// generic
#define  HAV_NONE      0
#define  HAV_DEFAULT  -1
// video codec IDs
#define  HAV_VIDEO_CODEC_RAW           1
#define  HAV_VIDEO_CODEC_FAST          2
#define  HAV_VIDEO_CODEC_BEST          3
#define  HAV_VIDEO_CODEC_BEST_PREDICT  4
// audio codec IDs
#define  HAV_AUDIO_CODEC_RAW   1
// divisor for fractional part of frame rate value
#define  HAV_FRAME_RATE_FRAC_DIVISOR  1000000000
// video depths
#define  HAV_VIDEO_BITS_8    8
#define  HAV_VIDEO_BITS_16  16
#define  HAV_VIDEO_BITS_24  24
// error consts
#define  E_HAV_OK       0
#define  E_HAV_FCREATE  1
#define  E_HAV_FWRITE   2
#define  E_HAV_MALLOC   3
#define  E_HAV_NODATA   4
#define  E_HAV_FORMAT   5
#define  E_HAV_MINVER   6
#define  E_HAV_FOPEN    7
#define  E_HAV_FREAD    8
#define  E_HAV_CMPS     9


typedef  Int32  HavError;
typedef  Uchar  HavPalette[HAV_PALETTE_ENTRIES][HAV_PALETTE_ENTRY_SIZE];

// headers used in HAV format

typedef struct {
  char    hdrID[4];
  Uint32  hdrSize;
  Uint32  fmtVer;
  Uint32  minVer;
  // video
  Int64   vFirstFPT;
  Uint32  vCodecId;
  Uint32  vFmtFlags;
  Uint32  vLen;
  Uint32  vResX;
  Uint32  vResY;
  Uint32  vBpp;
  Uint32  vFrameRateInt;
  Uint32  vFrameRateFrac;
  Uint32  vKeyFrameEvery;
  // audio
  Int64   aFirstFPT;
  Uint32  aCodecId;
  Uint32  aFmtFlags;
  Uint32  aLen;
  Uint32  aSampleRate;
  Uint32  aChannels;
  Uint32  aBitsChn;
  Uint32  aSamplesPerBlock;
} HavFileHeader;

typedef struct {
  Int64  filePos[HAV_FPT_ENTRIES];
  Int64  nextFPT;
} HavFPT;

typedef struct {
  char    packetId[4];
  Uint32  hdrSize;
	Uint32  hdrVer;
  Uint32  packetSize;
} HavHeader;

typedef struct {
  HavHeader  hdr;
  Uint32     dataSize;     // size of image data (palette included)
  Uint32     codecId;
  // following variables are valid only for 8-bit video
  Bool32     palPresent;   // nonzero = pallete follows image data
  Int64      palOffset;    // offset to palette packet
} HavVideoHeader;

typedef struct {
  HavHeader  hdr;
  Uint32     firstEntry;    // offset to the first color in palette
  Uint32     numOfEntries;  // number of colors
} HavPaletteHeader;

typedef struct {
  HavHeader  hdr;
  Uint32     dataSize;
  Uint32     codecId;
  Uint32     numSamples;
} HavAudioHeader;

typedef struct {
  HavFPT  FPT;     // current File Position Table
  int     FPTPos;  // current entry in File Position Table
  Int64   FPTFilePos;
  Bool    streamCreated;
  Uint    curFPT;    // current FPT in file (writing)
  Uint    curBlock;  // current block in file (block= every FPT member)
} HavStream;

typedef struct {
  HavFileHeader  hdr;
  // end of data from file
  FHandle     *f;  // FIXME: make separate handles for video and audio because of problems with multi-threading!
  HavStream   v;
  HavStream   a;
  double      vFrameRate;
  Uint32      vFrameSize;
  int         vSkipFrames;
  Pointer     vPrevImgDataP;
  HavPalette  *palP;  // current palette
  Uint32      palLocation;
  FOffset     palOffset;
  Bool        palRefreshed;
  int         aSampleSize;
  Uint32      aBlocks;
  Pointer     aBuf;
  int         aBufPos;
  int         aBufUsed;
  int         aSkipBytes;
} Hav;


// main functions
HavError  hav_open               (Hav *havP, const char *fileName);
void      hav_video_get_format   (Hav *havP, Uint32 *numFrames, Uint32 *codecId, Uint32 *resX, Uint32 *resY,
                                  Uint32 *bpp, double *frameRate, Int32 *keyFrameEvery);
void      hav_audio_get_format   (Hav *havP, Uint32 *numSamples, Uint32 *codecId, 
                                  Uint32 *sampleRate, Uint32 *channels, Uint32 *bitsChn);
void      hav_video_seek         (Hav *havP, Uint32 frameNum);
HavError  hav_video_read         (Hav *havP, Pointer imgDataP);
Bool32    hav_video_get_palette  (Hav *havP, Pointer palDataP, Bool *palRefreshed);
void      hav_audio_seek         (Hav *havP, Uint32 sample);
HavError  hav_audio_read         (Hav *havP, Pointer sndDataP, Uint32 numSamples, Uint32 *samplesOut);
HavError  hav_create             (Hav *havP, const char *fileName);
void      hav_video_set_format   (Hav *havP, Uint32 codecId, Uint32 resX, Uint32 resY, Uint32 bpp,
                                  double frameRate, Int32 keyFrameEvery);
void      hav_audio_set_format   (Hav *havP, Uint32 codecId, Uint32 samplesRate, 
                                  Uint32 channels, Uint32 bitsChn, Int32 samplesPerBlock);
HavError  hav_video_write_ex     (Hav *havP, Pointer imgDataP, Pointer palDataP, Uint32 *outputSizeP);
HavError  hav_video_write        (Hav *havP, Pointer imgDataP);
HavError  hav_audio_write_ex     (Hav *havP, Pointer sndDataP, Uint32 numSamples, Uint32 *outputSize);
HavError  hav_audio_write        (Hav *havP, Pointer sndDataP, Uint32 numSamples);
HavError  hav_close              (Hav *havP);

// extra
Hav       *hav_handle_alloc      (void);
void      hav_handle_free        (Hav *havP);
Uint32    hav_encode             (Uchar *cmpsDataP, Uchar *imgDataP, Uchar *prevImgDataP, Uint32 resX,
                                  Uint32 resY, Uint32 bpp, Uint32 codecID, Uint32 *trueCodecId);
HavError  hav_decode             (Uchar* imgDataP, Uchar* cmpsDataP,
                                  Uint32 resX, Uint32 resY, Uint32 bpp, Uint32 codecId);


#ifdef __cplusplus
}
#endif

#endif  // _HAV_H_
